﻿/*
    artifactInfoConfig = {
        options: { // optional, override production options
        }, 
        model: { // the wizzi model acting as artifact template
            src: '... relative or absolute file path' // path to the model
            schema: '...', // the schema that will load and validate the model
            globalValues: { // optional context values exposed during the evaluation of template and context models
                author: 'stefi',
                now: Date.now()
                ..
            },
            contexts: [ // wizzi models acting as data contexts during the evaluation of the template model
                {
                    src: '... relative or absolute file path' // path to the model
                    schema: '...', // the schema that will load and validate the model
                    contexts: [ // optional wizzi models acting as data contexts during the evaluation of the context model
                        ...
                    ],
                    coll: {
                        name: modelRef.collName, // property name of the collection
                        itemName: modelRef.collItemName, // name by which the collection item will be exported
                        collPathTemplateValues: [
                            pathTemplateValue: {
                                attribute: 'x', // attribute || function
                                token: 'y'
                            },
                            {
                                ....
                            }
                        ]
                    }
                },
                {
                    ...
                }
            ],
        },
        gen: { // generation info
            generator: '...'  // generator that will transform the wizzi model acting as artifact template.
        },
        dest: { // destination info for persistence
            path: 'app/app'
        }
    };
*/
var assert = require('assert');
var path = require('path');
var util = require('util');
var verify = require('../util/verify');
var interpolate = require('../util/interpolate');
var ModelInfo = require("../model/modelinfo");
var error = require("../util/error");
var log = require("../util/log")(module);

var ArtifactInfo = function (config) {
    this.name = config.name;
    this.options = config.options;
    if (verify.isObject(config.model) === false) {
        throw new Error("Artifact config " + confi.name + ' must have a model info');
    }
    this.modelInfo = new ModelInfo(config.model);
    this.genInfo = config.gen;
    this.destInfo = config.dest;
    this.genContexts = [];
};

// called by the generation manager
ArtifactInfo.prototype.initialize = function (productionManager) {
    this.___state = productionManager.___state;
    this.modelInfo.state(productionManager.___state);
    // merge production and artifact options
    var save = this.options || {};
    this.options = productionManager.options;
    for (key in save) {
        this.options[key] = save[key];
    }
    // WARN this works only after options merge
    // log.info('Initialized artifact: ' + this.toString());
}

// When a collection artifact add a genContext for a collection item
ArtifactInfo.prototype.addGenContext = function (ctx) {
    this.genContexts.push(ctx);
}

ArtifactInfo.prototype.isWizziModelArtifact = function () {
    return this.modelInfo.getContextCollection() === null &&
           verify.isObject(this.genInfo) &&
           verify.isString(this.genInfo.generator) &&
           this.genInfo.generator !== 'swig';
}

ArtifactInfo.prototype.isWizziCollectionArtifact = function () {
    return verify.isObject(this.modelInfo.getContextCollection()) &&
           verify.isObject(this.genInfo);
}

ArtifactInfo.prototype.isSwigArtifact = function () {
    return verify.isObject(this.genInfo) &&
           this.genInfo.generator === 'swig';
}

ArtifactInfo.prototype.isFinalArtifact = function () {
    return verify.isObject(this.genInfo) === false ||
           typeof this.genInfo.generator === 'undefined';
}

/*
    Returns a collection of infos for persisting to file
    the artifacts of a collection artifact.
    [
        {
            artifactInfo: ..., // the artifact config info (this object)
            genContext: ..., // contains the generated text
            filepath: ... // to filepath to write to
        }
    ]
*/
ArtifactInfo.prototype.getItemsToPersistToFile = function () {

    var _this = this;
    var result = [];

    this.genContexts.forEach(function (genContext) {

        var srcPath = genContext.srcPath; 

        if (_this.isWizziCollectionArtifact()) {

            var collItemInstance = genContext.model; // Required for resolving filepath.
            assert.strictEqual(verify.isObject(collItemInstance), true, 'genContext.model must contain an object.');
            var ipcontext = _this.getInterpolatePathNameContext(collItemInstance);
            var filepath = interpolate(_this.getFilepath(srcPath), ipcontext, { delimiter: '{}' })
            // console.log('ArtifactInfo.prototype.getCollectionContexts.filepath', filepath);
            var gc = {
                artifactInfo: _this,
                genContext: genContext,
                filepath: filepath
            };
            result.push(gc);

        } else {

            var filepath = _this.getFilepath(srcPath);
            // console.log('srcPath, filepath', srcPath, filepath)
            var gc = {
                artifactInfo: _this,
                genContext: genContext,
                filepath: filepath
            };
            result.push(gc);
        }

    });

    return result;

}

/*
    Input: an item of the model collection that drives the generation
           of this artifact.
*/
ArtifactInfo.prototype.getInterpolatePathNameContext = function (collItemInstance) {
    // console.log('ArtifactInfo.getInterpolatePathNameContext.collItemInstance', util.inspect(collItemInstance, { depth: 2 }));
    if (this.isWizziCollectionArtifact()) {
        var _this = this;
        var result = {};
        var contextCollInfo = this.modelInfo.getContextCollection();
        contextCollInfo.pathTemplateValues.forEach(function (item) {
            if (item.function)
                result[item.token] = collItemInstance[item.function]();
            else if (item.attribute)
                result[item.token] = collItemInstance[item.attribute];
            else
                error('path template value must contain an attribute or a function value, token: ' + item.token);
        });
        // console.log('ArtifactInfo.getInterpolatePathNameContext.result', util.inspect(result, { depth: 2 }));
        return result;
    } else {
        this.error("Method getInterpolatePathNameContext called on an artifact that is not a wizzi collection artifact.");
    }
}

ArtifactInfo.prototype.getFilepath = function (srcPath) {
    var opt = this.options;
    var dest = this.destInfo;
    var msg = ['ArtifactInfo.getFilepath', 'config.dest'];

    if (!verify.isObject(dest)) this.error('ArtifactInfo.' + msg[0] + ' require a ' + msg[1] + 'object.');
        if (verify.isString(dest.fullpath)) {
        return dest.fullpath; // ok its enough, dest contains a fullpath
    }

    if (!verify.isString(dest.path)) this.error('ArtifactInfo.' + msg[0] + ' require a ' + msg[1] + '.path value of type string.');

    srcPath = dest.srcReplace ? dest.srcReplace : srcPath;

    if (!verify.isString(srcPath)) this.error('ArtifactInfo.' + msg[0] + ' require a srcPath of type string.');

    var destpath = path.join(dest.path, srcPath);
    
    if (this.isFinalArtifact() === false) {
        var ext = dest.extension || '.js'; // hardcoded default
        ext = verify.startsWith(ext, '.') ? ext : '.' + ext;
        if (ext.toLowerCase() !== '.ittf') {
            if (destpath.substr(-5, 5).toLowerCase() === '.ittf') {
                destpath = destpath.substr(0, destpath.length - 5);
            }
        }
        ext = verify.endsWith(destpath, ext) ? '' : opt.dotgExtensionPrefix ? '.g' + ext : ext;
        destpath = destpath + ext;
    }
    
    if (verify.isAbsolutePath(destpath)) return destpath;

    if (!verify.isString(opt.basedir)) this.error('When config.dest.path is a relative path ArtifactInfo.' + msg[0] + ' require a ' + msg[1] + '.basedir value of type string.');
    return path.join(opt.basedir, destpath);
}

ArtifactInfo.prototype.error = function (message) {
    throw new Error(message + '\nArtifact:' + this.toString());
}

ArtifactInfo.prototype.dump = function () {
    throw new Error('Not implemented');
}

ArtifactInfo.prototype.relPath = function () {
    return this.modelInfo.srcFullPath().substr(this.options.basedir.length + 1);
}

ArtifactInfo.prototype.toString = function () {
    // console.log('ArtifactInfo.options', util.inspect(this.options, { depth: 1 }));
    return [
        'model.src:', this.relPath(),
        'gen.generator:', this.genInfo ? this.genInfo.generator : ''
    ].join(' ');
}

ArtifactInfo.isArtifactConfig = function (test) {
    if (!test) return false;
    if (!verify.isObject(test.options)) return false;
    if (!verify.isObject(test.model)) return false;
    if (!verify.isObject(test.dest)) return false;
    if (!verify.isObject(test.gen)) return false;
    return true;
}

module.exports = ArtifactInfo;
