﻿/*
    Async artifact generation runner.
    Input: an artifact config info (constructor injected)
    Output: the artifact config info enriched with (one or many) executed genContext(s), 
            ready to be persisted (callback ok result)
*/
var assert = require('assert');
var path = require('path');
var util = require('util');
var swig = require('swig');
var verify = require("../util/verify");
var log = require("../util/log")(module);
var file = require("../util/file");
var AsyncModel = require('../model/loader');
var GenContext = require('../artifact/gencontext');

/*
    A runner instance operates on an a single artifactInfo
*/
var Runner = function (artifactInfo) {
    this.artifactInfo = artifactInfo;
    log.setLevel(artifactInfo.options.verbose || 0);
};

Runner.prototype.run = function (callback) {
    var _this = this;
    var modelInfo = this.artifactInfo.modelInfo; // artifactConfig.model property
    modelInfo.___state = this.artifactInfo.___state; // expose production state on modelInfo
    var genInfo = this.artifactInfo.genInfo; // artifactConfig.gen property
    if (this.artifactInfo.isWizziModelArtifact()) {
        this.runWizziModelArtifact(modelInfo, genInfo, function (err, result) {
            callback(err, result);
        })
    } else if (this.artifactInfo.isWizziCollectionArtifact()) {
        this.runWizziCollectionArtifact(modelInfo, genInfo, function (err, result) {
            callback(err, result);
        })
    } else if (this.artifactInfo.isSwigArtifact()) {
        this.runSwigArtifact(modelInfo, function (err, result) {
            callback(err, result);
        })
    } else if (this.artifactInfo.isFinalArtifact()) {
        this.runFinalArtifact(modelInfo, function (err, result) {
            callback(err, result);
        })
    } else {
        throw new Error('Artifact is in an invalid state: ' + util.inspect(this.artifactInfo, { depth: 2 }));
    }
}

Runner.prototype.runWizziModelArtifact = function (modelInfo, genInfo, callback) {
    log.warn('runWizziModelArtifact ' + this.artifactInfo.name);
    var _this = this;

    // set modelInfos
    var modelInfos;
    if (modelInfo.isDirectory()) {
        modelInfos = modelInfo.getModelInfos();
    } else {
        modelInfos = [modelInfo];
    }

    // retrieve generator
    var generator = modelInfo.___state.pman.getArtifactGenerator(genInfo.generator);
    if (generator == null) {
        callback('Generator not found: ' + genInfo.generator + ', for artifact ' + _this.artifactInfo.toString(), null);
    }

    AsyncModel.loadMany(modelInfos, function (err, modelObjects) {
        if (err) { callback(err, null); return; }

        for (var i = 0; i < modelInfos.length; i++)
        {
            var mInfo = modelInfos[i];
            var mObject = modelObjects[i];

            var genContext = new GenContext({
                model: null,
                srcPath: mInfo.src(),
                srcFullPath: mInfo.srcFullPath(),
                options: _this.artifactInfo.options
            });

            new generator.gen(mObject, genContext);
            _this.artifactInfo.addGenContext(genContext); // augment artifactInfo with result

        }
        callback(null, _this.artifactInfo);
    });

}

Runner.prototype.runWizziCollectionArtifact = function (modelInfo, genInfo, callback) {
    log.warn('runWizziCollectionArtifact ' + this.artifactInfo.name);
    var _this = this;
    // modelInfo describes a wizzi model acting as an artifact template, load it async

    // modelInfo.src is a file
    AsyncModel.load(modelInfo, function (err, modelObjects) {
        if (err) { callback(err, null); return; }
        log.warn('collection artifact: ', _this.artifactInfo.toString());
        var generator = modelInfo.___state.pman.getArtifactGenerator(genInfo.generator);
        if (generator == null) {
            callback('Generator not found: ' + genInfo.generator + ', for artifact ' + _this.artifactInfo.toString(), null);
        }
        assert.strictEqual(verify.isArray(modelObjects), true, 'modelObjects must be an array when artifact is a collection artifact.');
        // modelObjects is a collection of resolved template artifacts
        modelObjects.forEach(function (item) {
            var genContext = new GenContext({
                model: item.___collItem, // required by collection artifacts for resolving the destination filepath.
                srcPath: modelInfo.src(),
                srcFullPath: modelInfo.srcFullPath(),
                options: _this.artifactInfo.options
            });
            new generator.gen(item, genContext); // this is the big job !!!
            _this.artifactInfo.addGenContext(genContext); // augment artifactInfo with result
        });
        callback(null, _this.artifactInfo);
    });

}

Runner.prototype.runSwigArtifact = function (modelInfo, callback) {
    log.warn('runSwigArtifact ' + this.artifactInfo.name);
    var _this = this;

    // Load the contextModelObjects for the swig template transformation
    AsyncModel.loadContexts(modelInfo, function (err, contextModelObjects) {
        if (err) { callback(err, null); return; }
        // create the template context object
        var context = {};
        contextModelObjects.forEach(function (item) {
            context[item.___exportName] = item;
        });

        if (modelInfo.isDirectory()) {

            var sources = modelInfo.getSources();
            sources.forEach(function (item) {

                // transform the source (template)
                var transformed = swig.render(item.content, { locals: context });

                // write the result into a new GenContext
                var genContext = new GenContext({
                    model: null,
                    srcPath: item.relpath,
                    srcFullPath: item.fullpath,
                    options: _this.artifactInfo.options
                });
                genContext.write(transformed);
                _this.artifactInfo.addGenContext(genContext); // augment artifactInfo with result
            });

        } else {

            // read the template content
            var template = modelInfo.getSource();
            // transform the template
            var transformed = swig.render(template, { locals: context });

            // write the result into a new GenContext
            var genContext = new GenContext({
                model: null,
                srcPath: modelInfo.src(),
                srcFullPath: modelInfo.srcFullPath(),
                options: _this.artifactInfo.options
            });
            genContext.write(transformed);
            _this.artifactInfo.addGenContext(genContext); // augment artifactInfo with result
        }

        callback(null, _this.artifactInfo);

    });
}

/*
    A final artifact is read from disk and written into a genContext
    without template transformation and generation.
    The source can be a folder, in this case a genContext is created
    for each file in the folder.
*/
Runner.prototype.runFinalArtifact = function (modelInfo, callback) {
    log.warn('runFinalArtifact ' + this.artifactInfo.name);

    var _this = this;

    if (modelInfo.isDirectory()) {

        var files = modelInfo.getFiles();
        files.forEach(function (fileInfo) {

            // console.log('Runner.runFinalArtifact.item', modelInfo.src(), item.filepath);
            var genContext = new GenContext({
                model: null,
                srcPath: fileInfo.relpath,
                srcFullPath: fileInfo.fullpath,
                options: _this.artifactInfo.options
            });
            genContext.writeFile(fileInfo.fullpath);
            _this.artifactInfo.addGenContext(genContext); // augment artifactInfo with result
        });

    } else {

        var genContext = new GenContext({
            model: null,
            srcPath: modelInfo.src(),
            srcFullPath: modelInfo.srcFullPath(),
            options: _this.artifactInfo.options
        });
        genContext.write(modelInfo.getSource());
        _this.artifactInfo.addGenContext(genContext);

    }
    callback(null, this.artifactInfo);
}

module.exports = Runner;
